#include "../Core_Mesh/Iterators.h"
#include "DisplayWindow.h"

using namespace XMeshLib;

//=================================================
//==================== Overide ====================
//=================================================
int DisplayWindow::handle(int event)
{
	int button;
	int x, y;
	switch(event) {
	case FL_PUSH:
		button = Fl::event_button();
		x = Fl::event_x();
		y = Fl::event_y();
		MouseDown(button, x, y);
		return 1;
	case FL_DRAG:
		button = Fl::event_button();
		x = Fl::event_x();
		y = Fl::event_y();
		MouseMotion(button, x, y);
		return 1;
	case FL_RELEASE:   
		button = Fl::event_button();
		x = Fl::event_x();
		y = Fl::event_y();
		MouseUp(button, x, y);
		return 1;
	case FL_FOCUS :
	case FL_UNFOCUS :
//    ... Return 1 if you want keyboard events, 0 otherwise
    return 1;
	case FL_KEYBOARD:
//    ... keypress, key is in Fl::event_key(), ascii in Fl::event_text()
//    ... Return 1 if you understand/use the keyboard event, 0 otherwise...
	case FL_SHORTCUT:
//   ... shortcut, key is in Fl::event_key(), ascii in Fl::event_text()
//    ... Return 1 if you understand/use the shortcut event, 0 otherwise...
    return 1;
	default:
    // pass other events to the base class...
	return Fl_Gl_Window::handle(event);
  }
}

// Reshape function
void DisplayWindow::Reshape(GLsizei w, GLsizei h)
{
	glViewport(0, 0, w, h);

	win_width = w;
	win_height = h;
}

void DisplayWindow::ComputeBoundingBox()
{
	if (!theMesh)
		return;

	double minx, maxx, miny, maxy, minz, maxz;
	minx = miny = minz = 1e10;
	maxx = maxy = maxz = -1e10;

	for (MeshVertexIterator viter(theMesh); !viter.end(); ++viter)
	{
		Vertex *v = *viter;
		if (v->point()[0] < minx) minx = v->point()[0];
		if (v->point()[0] > maxx) maxx = v->point()[0];

		if (v->point()[1] < miny) miny = v->point()[1];
		if (v->point()[1] > maxy) maxy = v->point()[1];

		if (v->point()[2] < minz) minz = v->point()[2];
		if (v->point()[2] > maxz) maxz = v->point()[2];
	}

	m_bbox.minX() = minx; m_bbox.maxX() = maxx;
	m_bbox.minY() = miny; m_bbox.maxY() = maxy;
	m_bbox.minZ() = minz, m_bbox.maxZ() = maxz;

	//////////////////////////
	Vector3f homepos, homerotaxis;
	homepos.x = (m_bbox.minX() + m_bbox.maxX())/2.0;//0;
	homepos.y = (m_bbox.minY() + m_bbox.maxY())/2.0;//0;
	homepos.z = m_bbox.diag()+(m_bbox.minZ() + m_bbox.maxZ())/2.0;//3.0;
	
	homerotaxis = Vector3f(0, 0, 1);

	theCamera.SetHome(homepos, 0, homerotaxis, DEFAULT_FOV);
	rotate_dist = view_dist = m_bbox.diag();

	theCamera.GoHome();
}


//Mouse and Keyboard handling
void DisplayWindow::MouseDown(int button, int x, int y)
{
	float nx, ny;
	Mousei2f(x, y, &nx, &ny);
		
	last_mouse_x = nx;
	last_mouse_y = ny;
	last_button = button;

	switch (button) 
	{
	case FL_LEFT_MOUSE:
		if (Fl::event_shift())
		{
			theCamera.status = Camera::ZOOMING;
			cursor(FL_CURSOR_NS);
		}
		else if (Fl::event_ctrl())
		{
			theCamera.status = Camera::PANNING;
			cursor(FL_CURSOR_WE);
		}
		else if (Fl::event_alt())
		{
			theCamera.status = Camera::ROTATING;
			cursor(FL_CURSOR_NESW);
		}
		else
			theCamera.status = Camera::IDLING;
        

		if (enable_picking)
		{
			DoPicking(x, y, picking_mode);
			cursor(FL_CURSOR_CROSS);
		}

		break;
    case FL_MIDDLE_MOUSE:
		break;
    case FL_RIGHT_MOUSE:
		break;
    default:
        break;
	}	
}


void DisplayWindow::MouseUp(int button, int x, int y)
{
}


void DisplayWindow::MouseMotion(int button, int x, int y)
{
	float cur_mouse_x, cur_mouse_y;
	float delta_x, delta_y;

	Mousei2f(x, y, &cur_mouse_x, &cur_mouse_y);

	delta_x = cur_mouse_x - last_mouse_x;
	delta_y = cur_mouse_y - last_mouse_y;

	if (last_button == FL_LEFT_MOUSE)
	{
		switch (theCamera.status)
		{
		case Camera::PANNING:
			Translate(delta_x, delta_y, 0);
			break;
		case Camera::ROTATING:
			Rotate(last_mouse_x, last_mouse_y, cur_mouse_x, cur_mouse_y);
			break;
		case Camera::ZOOMING:
			Translate(0, 0, delta_y);
			break;
		case Camera::IDLING:
		default:
			break;
		}
	}

	last_mouse_x = cur_mouse_x;
	last_mouse_y = cur_mouse_y;

	redraw();
}

// Helper function - convert integer coordinates of mouse position to
// normalized floats
void DisplayWindow::Mousei2f(int x, int y, float *xf, float *yf)
{
	int screenwidth = Screenw();
	int screenheight = Screenh();

	float r = (float)screenwidth/screenheight;
	if (r > 1.0) {
		*xf = r * (2.0f * x / screenwidth - 1.0f);
		*yf = 1.0f - 2.0f * y / screenheight;
	} else {
		*xf = 2.0f * x / screenwidth - 1.0f;
		*yf = (1.0f - 2.0f * y / screenheight) / r;
	}
}

// Mouse event helper function - rotate
void DisplayWindow::Rotate(float ox, float oy, float nx, float ny)
{
	Quatf q;
	Camera::Arc2Quaternion(ox, oy, nx, ny, q);

	if (rotate_dist == 0) {
		q[1] = -q[1];
		q[2] = -q[2];
	}

	theCamera.Rotate(q, rotate_dist);
}


// Mouse event helper function - translate
void DisplayWindow::Translate(float dx, float dy, float dz)
{
	// some auto param ajustment to be refined later
	float scalefactor = 0.5f*view_dist;
	dx *= scalefactor*theCamera.fov;
	dy *= scalefactor*theCamera.fov;
	dz = view_dist * (exp(-0.5f * dz) - 1.0f);

	// moving the camera in the opposite direction to
	// the direction we want the object to move
	theCamera.Move(-dx, -dy, -dz);
	UpdateDist(-dx, -dy, -dz);
}


// Mouse event helper function - update distance
void DisplayWindow::UpdateDist(float dx, float dy, float dz)
{
	Vector3f viewdir(0, 0, -view_dist);
	Vector3f delta(-dx, -dy, -dz);

	Vector3f newviewdir = viewdir + delta;

	view_dist = -newviewdir.z;
	// Always rotate around the center of the object
	// which is assumed to be place at the origin of the world coordinate
	rotate_dist = view_dist;

}

// (3) Picking control functions
void DisplayWindow::ClearPicks()
{
	switch (picking_mode)
	{
	case PICKING_VERTEX:
		pickedVs.clear();
		break;
	case PICKING_EDGE:
		pickedEs.clear();
		break;
	case PICKING_FACE:
		pickedFs.clear();
		break;
	default:
		assert(0);
	}
}

void DisplayWindow::DoPicking(int x, int y, int pmode)
{
#define DEBUG_DOPICKING	0
	GLint hits;
	GLint viewport[4];
	GLuint selectBuf[SELBUF_SIZE];
	glGetIntegerv(GL_VIEWPORT, viewport);
	glSelectBuffer(SELBUF_SIZE, selectBuf);
	glRenderMode(GL_SELECT);
	glInitNames();
	glPushName(0);
	glPushMatrix();
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPickMatrix((GLdouble)x, (GLdouble)(viewport[3] - y), 5, 5, viewport);
	theCamera.Projection(view_dist/DOF, view_dist*DOF, win_width, win_height);
	glMatrixMode(GL_MODELVIEW);
	theCamera.Modelview();
	// currently only picking on point set supported
	RenderMesh4Picking(pmode);
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glFlush();
	hits = glRenderMode(GL_RENDER);
	if (hits < 0)
	{
		fprintf(stderr, "Error: Selection Buffer Overflow!\n");
		return;
	}
#if DEBUG_DOPICKING
	printf("hits: %d\n", hits);
#endif
	ProcessHits(hits, selectBuf, pmode);
	redraw();
#undef DEBUG_DOPICKING
}


void DisplayWindow::ProcessHits(GLint hits, GLuint buffer[], int pmode)
{
	int i, j;
	GLuint names, *ptr;
	std::vector<std::pair<GLfloat, GLuint>> name_array;	//each vertex/face has an "id" name, set during its rendering
	std::vector<std::pair<GLfloat, GLuint>>::iterator niter;
	std::vector<std::pair<GLfloat, std::pair<GLuint, GLuint>>> edge_array;
	std::vector<std::pair<GLfloat, std::pair<GLuint, GLuint>>>::iterator eiter;
	
	std::pair<GLfloat, GLuint> name_pair;
	std::pair<GLfloat, GLuint> other_pair;
	std::pair<GLfloat, std::pair<GLuint, GLuint>> edge_pair;

	ptr = (GLuint *) buffer;
	
	for (i = 0; i < hits; ++i)
	{
		names = *ptr;
		ptr += 3;

		for (j = 0; j < names; ++j)
		{
			// In vertex picking mode, we will have to eliminate duplicates
			// Duplicates don't happen in face/edge mode
			// !!This is related to the way we render the model for picking
			name_pair = std::make_pair((GLfloat)*(ptr-2)/0x7fffffff, (*ptr));
			name_array.push_back(name_pair);

#if DEBUG_PROCESSHITS
			fprintf(stderr, "name: %u z1: %g z2: %g\n", *ptr, (float)*(ptr-2)/0x7fffffff, (float)*(ptr-1)/0x7fffffff);
#endif				
			ptr++;
		}
	}

	// A minor problem: if two points are collinear they may both be picked
	// we need to throw away the one that is farther away
	// A tricky problem: the point which we don't actually see (hidden further away)
	// may gets picked if the picking is not accurate enough and does not actually 
	// pick the point in the front which we DO see
	// Solution: sort the names according to their z1 value
	switch(pmode)
	{
	case PICKING_FACE:
	case PICKING_VERTEX:
		if (name_array.size() > 1)
		{
			std::sort(name_array.begin(), name_array.end());
		}
		break;
	case PICKING_EDGE:
		for (niter = name_array.begin(); niter != name_array.end(); ++niter)
		{
			name_pair = (*niter);
			++niter;
			other_pair = (*niter);			
			GLfloat z1min = __min(name_pair.first, other_pair.first);
			edge_array.push_back(std::make_pair(z1min, std::make_pair(name_pair.second, other_pair.second)));
		}
		if (edge_array.size() > 1)
		{
			std::sort(edge_array.begin(), edge_array.end());
		}
		break;
	default:
		break;
	}

	if (name_array.empty())
	{
#if DEBUG_PROCESSHITS
		fprintf(stderr, "Name array empty!\n");
#endif
		return;
	}
		
	MeshFaceIterator fiter(theMesh);
	MeshVertexIterator viter(theMesh);

	switch (pmode)
	{
	// process each name
	case PICKING_FACE:		
		for (niter = name_array.begin(); niter != name_array.end(); ++niter)
		{
			name_pair = *niter;
			Face *f = theMesh->idFace(name_pair.second);
			if (f)
			{
#if DEBUG_PROCESSHITS
				fprintf(stdout, "Face %d picked!\n", f->id());
#endif
				f->sharp() = !f->sharp();
				std::vector<Face *>::iterator fit = std::find(pickedFs.begin(), pickedFs.end(), f);
				if (fit==pickedFs.end())
					pickedFs.push_back(f);
				else
					pickedFs.erase(fit);
			}
#if DEBUG_PROCESSHITS
			else
				fprintf(stderr, "Face not found!\n");
#endif
			break;	
		}
		break;
	case PICKING_VERTEX:
		for (niter = name_array.begin(); niter != name_array.end(); ++niter)
		{
			name_pair = *niter;
			Vertex *v = theMesh->idVertex(name_pair.second);

			if (v)
			{
#if DEBUG_PROCESSHITS
				fprintf(stdout, "Vertex %d picked!\n", v->id());
#endif			
				v->sharp() = !v->sharp();
				std::vector<Vertex *>::iterator vit = std::find(pickedVs.begin(), pickedVs.end(), v);
				if (vit==pickedVs.end())
					pickedVs.push_back(v);
				else
					pickedVs.erase(vit);
			}
#if DEBUG_PROCESSHITS
			else
				fprintf(stderr, "Vertex not found!\n");
#endif
			break;
		}		
		break;
	case PICKING_EDGE:
		for (eiter = edge_array.begin(); eiter != edge_array.end(); ++eiter)
		{
			edge_pair = *eiter;
			Vertex *v0 = theMesh->idVertex(edge_pair.second.first);
			Vertex *v1 = theMesh->idVertex(edge_pair.second.second);
			assert(v0);
			assert(v1);
			Edge *e = theMesh->vertexEdge(v0, v1);
			if (e)
			{
#if DEBUG_PROCESSHITS
				fprintf(stdout, "Edge (%d,%d) picked!\n", e->vertex(0), e->vertex(1));
#endif
				//fprintf(stdout, "Edge (%d,%d) picked!\n", e->vid(0), e->vid(1));
				e->sharp() = !e->sharp();
				std::vector<Edge *>::iterator eit = std::find(pickedEs.begin(), pickedEs.end(), e);
				if (eit==pickedEs.end())
					pickedEs.push_back(e);
				else
					pickedEs.erase(eit);
			}
#if DEBUG_PROCESSHITS
			else
				fprintf(stderr, "Edge not found!\n");
#endif
			break;
		}
		break;
	default:
		break;
	}
#undef DEBUG_PROCESSHITS
}
