#ifndef _XMESHLIB_MUNKRES_H_
#define _XMESHLIB_MUNKRES_H_

#include <vector>
#include "../Core_Mesh/Point.h"

namespace XMeshLib
{
	class Munkres
	{
	public:
		Munkres(int _n, int _m);
		//Inputing dimensions of the matrix: _n <= _m

		~Munkres()
		{
			for (int i=0;i<n;++i)
				delete [] Costs[i];
			delete [] Costs;
		}

		void SetInitMatrix(double ** _c)
		{
			Costs = new double * [n];
			for (int i=0;i<n;++i)
			{
				Costs[i] = new double [m];
				for (int j=0;j<m;++j)
				{
					Costs[i][j] = _c[i][j];
				}
			}
		}

		void OutputMatrix();

		void static VectorsDist2Matrix(std::vector<Point> vlist1, std::vector<Point> vlist2,
						int & row, int & col, double ** & outMatrix);
		
		double EstimateFinalCosts(double ** mat);

		void static FreeMatrix(double ** m, int row);

		void Process();

		typedef std::pair<int, int> IntPair;
		std::vector<IntPair> OptMatch;	

	private:
		void Step_1();
		void Step_2();
		void Step_3();
		void Step_4();
		void Step_5();
		void Step_6();

		//for step 4
		void find_a_zero(int & row, int & col);
		bool star_in_row(int row);
		int find_star_in_row(const int row);

		//for step 5
		int find_star_in_col(const int col);
		int find_prime_in_row(const int row);
		void convert_path(int count);
		void clear_covers();
		void erase_primes();

		//for step 6
		double find_smallest();


	private:
		int n;
		int m;
		int min_m_n;
		double ** Costs;
		int step;
		int * R_cover;
		int * C_cover;
		int ** Mask;
		//int count;

		int Z0_r, Z0_c;	

		std::vector<int> path1;
		std::vector<int> path2;

	};
}

#endif
